use std::collections::VecDeque;
use std::fmt::Formatter;
use std::{cmp, fmt};

fn capitalise(s: &str) -> String {
    let mut c = s.chars();
    match c.next() {
        None => String::new(),
        Some(f) => f.to_uppercase().collect::<String>() + c.as_str(),
    }
}

/// A basic register capable of storing binary data.
#[derive(Debug)]
pub struct BasicRegister {
    /// A [Vec] that stores the binary data of the register.
    memory: VecDeque<bool>,
}

impl BasicRegister {
    pub fn adjusted_by_size(&mut self, resulting_size: usize) -> Self {
        let current_memory_size: usize = self.memory.len();
        let difference: i32 = current_memory_size as i32 - resulting_size as i32;

        match current_memory_size > resulting_size {
            true => BasicRegister::new(
                (difference as usize..current_memory_size)
                    .map(|i| self.memory[i])
                    .collect::<VecDeque<bool>>(),
            ),
            false => {
                let mut resulting_memory: VecDeque<bool> = vec![false; -difference as usize].into();
                resulting_memory.append(&mut self.memory);
                BasicRegister::new(resulting_memory)
            }
        }
    }

    pub fn negate(&mut self) {
        self.memory = self.memory.iter().map(|val| !val).collect();
    }

    pub fn negated(&self) -> Self {
        BasicRegister::new(self.memory.iter().map(|val| !val).collect())
    }

    pub fn left_shift(
        &mut self,
        shift_in_value_default: Option<bool>,
        bits_shifted_default: Option<i32>,
    ) -> VecDeque<bool> {
        let shift_in_value: bool = shift_in_value_default.unwrap_or(false);
        let bits_shifted: usize = bits_shifted_default.unwrap_or(1) as usize;

        let mut shifted_bits: VecDeque<bool> = [].into();
        (0..bits_shifted).for_each(|_| shifted_bits.push_back(self.memory.pop_front().unwrap()));
        (0..bits_shifted).for_each(|_| self.memory.push_back(shift_in_value));

        shifted_bits
    }

    pub fn right_shift(
        &mut self,
        shift_in_value_default: Option<bool>,
        bits_shifted_default: Option<i32>,
    ) -> VecDeque<bool> {
        let shift_in_value: bool = shift_in_value_default.unwrap_or(false);
        let bits_shifted: usize = bits_shifted_default.unwrap_or(1) as usize;

        let mut shifted_bits: VecDeque<bool> = [].into();
        (0..bits_shifted).for_each(|_| shifted_bits.push_back(self.memory.pop_back().unwrap()));
        (0..bits_shifted).for_each(|_| self.memory.push_front(shift_in_value));

        shifted_bits
    }

    pub fn new(memory: VecDeque<bool>) -> Self {
        Self { memory }
    }

    pub fn len(&self) -> usize {
        self.memory.len()
    }
}

impl Default for BasicRegister {
    fn default() -> Self {
        Self {
            memory: [false].into(),
        }
    }
}

impl fmt::Display for BasicRegister {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        write!(f, "Memory: [")?;

        for (count, v) in self.memory.iter().enumerate() {
            if count != 0 {
                write!(f, ", ")?;
            }
            write!(f, "{}", *v as u8)?;
        }

        write!(f, "]")
    }
}

impl fmt::Binary for BasicRegister {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        write!(f, "Memory: ")?;

        for v in self.memory.iter() {
            write!(f, "{}", *v as u8)?;
        }

        write!(f, "")
    }
}

fn u8_to_bool(number: u8) -> bool {
    number > 0
}

fn read_vec(variable_name: &str) -> Vec<char> {
    loop {
        print!("Enter {variable_name}: ");
        let input: String = text_io::read!("{}\n");
        match input.chars().all(|c| c == '1' || c == '0') {
            true => return input.chars().collect(),
            false => eprintln!(
                "[ERROR] {} may contain only 1-s and 0-s with no whitespaces!",
                capitalise(variable_name)
            ),
        }
    }
}

pub fn char_to_bool_vector(char_vector: Vec<char>) -> VecDeque<bool> {
    let mut bool_vector: VecDeque<bool> = VecDeque::new();

    for value in char_vector.iter() {
        match value {
            '0' => bool_vector.push_back(false),
            '1' => bool_vector.push_back(true),
            _ => (),
        }
    }

    bool_vector
}

pub fn get_memory(variable_name: &str) -> VecDeque<bool> {
    char_to_bool_vector(read_vec(variable_name))
}

pub fn align_registers(
    mut first_register: BasicRegister,
    mut second_register: BasicRegister,
) -> (BasicRegister, BasicRegister) {
    let first_size: usize = first_register.memory.len();
    let second_size: usize = second_register.memory.len();
    let required_size: usize = cmp::max(first_size, second_size);

    (
        first_register.adjusted_by_size(required_size),
        second_register.adjusted_by_size(required_size),
    )
}

fn binary_sum_with_carry(
    first_term: &BasicRegister,
    second_term: &BasicRegister,
) -> (BasicRegister, bool) {
    let mut sum: BasicRegister = BasicRegister::new(vec![false; first_term.len()].into());

    let mut carry: bool = false;
    let mut current_bit_sum: u8;
    for i in (1..first_term.len()).rev() {
        current_bit_sum = first_term.memory[i] as u8 + second_term.memory[i] as u8 + carry as u8;
        carry = u8_to_bool(current_bit_sum & 2);
        sum.memory[i] = u8_to_bool(current_bit_sum & 1);
    }

    let final_bit_sum: u8 = first_term.memory[0] as u8 + second_term.memory[0] as u8 + carry as u8;
    let final_carry: bool = u8_to_bool(final_bit_sum & 2);

    sum.memory[0] = u8_to_bool(final_bit_sum & 1);

    (sum, final_carry)
}

pub fn binary_sum(first_term: &BasicRegister, second_term: &BasicRegister) -> BasicRegister {
    binary_sum_with_carry(first_term, second_term).0
}

pub fn binary_subtraction(minuend: &BasicRegister, subtrahend: &BasicRegister) -> BasicRegister {
    let (difference, final_carry) = binary_sum_with_carry(minuend, &subtrahend.negated());

    match final_carry {
        true => {
            let mut complement = vec![false; difference.len() - 1];
            complement.append(&mut vec![true]);
            binary_sum(&difference, &BasicRegister::new(complement.into()))
        }
        false => difference.negated(),
    }
    // BasicRegister::default()
}
